#ifndef bqtRockEditLEVElHH
#define bqtRockEditLEVElHH

#include <algorithm>

#include "defs.hh"
#include "atoms.hh"

/* Monsters indexes:

    [n] = which position it goes into (1-5)
 
    [3] BE: ?
    [3] B2: Lift with track (*--/)
    AE: ?
    CA: ?
    [1] 96,9A,9E: Green guardian (G)
    
    (automatic) 40,44,48,4C,50,54,58,5C = megaman and regular powerups
    
    [1] 60,64,68,6C: Cutman (verify)  (also 46?)
    [1] 70,74,78,7C: Gutsman          (also F9?)
    [1] 80,84,88,8C: Bombman (verify) (also 4A?)
    [1] 90,94,98:    Fireman          (also 5A,42?)
    [1] A0,A4,A8,AC: Iceman (verify)  (also 46?)
    [1] B0,B4,B8,BC: Elecman (verify) (also 52?)
    [1] DF,E3,E3,E3,E3: Wily1
    [1] 63,46,4A,42,52? Clone robot?
    [1] 4A,6A,6E,76,DA? Wily3?
    [2] EA,EE,F2,FB,FF: Wily4?


 Others are in objects.cc
*/


struct Level
{
////////////////////////
    // PROPERTIES
///////////////////////    

    unsigned LevelNumber;
    const char* Name;
    
    Screen rooms[MaxRooms];
    
    /* Palettes:
         Pal0 [4]
         Pal1 [4]
         Pal2 [4]
         Pal3 [4]
         
         Megaman body [4] (also Weaponbar)
         Energyfill [4] (also Megaman face)
         Enemy1 [4]
         Enemy2 [4]
         
         Boss room blinks- Pal0,Pal1,Pal2,Pal3
    */
         
    

    unsigned char Palette[16*3];
    unsigned char AfterDoorPalette[16];
    
    Active actives[MaxActives];
    Enemy  enemies[MaxEnemies];
    
    LevelGraphicsData   levelgfx[MaxLevelGfx];
    MonsterGraphicsData monstergfx[MaxMonsterGfx];
    
    unsigned char ArrivalInput;

    /* Each level has different representations
     * in text form.
     */
    void (*CreateAscii)(Screen& screen);
    void (*CreateChar)(Screen& screen);
    
////////////////////////
    // METHODS
///////////////////////    

    void WritePalette(unsigned addr, unsigned addr_afterdoor) const;
    void WriteActives(unsigned first_screen, unsigned addr, unsigned end_addr) const;
    void WriteEnemies(unsigned first_screen,
                      unsigned ptraddr, unsigned& addr, unsigned end_addr) const;
    void WriteGFX(unsigned level_addr,unsigned index_addr,unsigned table_addr);

    void ReadPalette(unsigned addr, unsigned addr_afterdoor);
    void ReadActives(unsigned first_screen, unsigned addr);
    void ReadEnemies(unsigned first_screen, unsigned ptraddr);
    void ReadGFX(unsigned level_addr,unsigned index_addr,unsigned table_addr);
    
    /* Call ReadGFX only after Screens have been constructed, and only once. */
    
    void SortActives()
    {
        std::sort(actives, actives+MaxActives);
    }
    void SortEnemies()
    {
        std::sort(enemies, enemies+MaxEnemies);
    }
    

/*
Buzzers in Iceman level.
69
83
88
8B
A0
BA
D8
*/


    void AnalyzeObjects()
    {
        /* To be called after everything has been read from the ROM
         * and the ascii map has been generated, before dumping them
         * to output.
         */
        
        /* Delete actives that can be expressed on the map. */
        DeleteGutsBlockActives();
        
        /* Move the shutters from the active list to the object list. */
        MakeAsciiActives();
        
        /* Move movable enemies to the ascii map. */
        MakeAsciiEnemies();
        
        /* Last, sort the arrays. */
        SortActives();
        SortEnemies();
    }
    
    void ComposeObjects()
    {
        CreateGutsBlockActives();
        ReadAsciiActives();
        ReadAsciiEnemies();
        
        /* Last, sort the arrays. */
        SortActives();
        SortEnemies();
    }    
    
    void DeleteGutsBlockActives();
    void MakeAsciiActives();
    
    void MakeAsciiEnemies();

    void CreateGutsBlockActives();

    void ReadAsciiActives();
    void ReadAsciiEnemies();
    
    void AddActive(const Active& b)
    {
        for(unsigned a=0; a<MaxActives; ++a)
            if(actives[a].screen == 255)
            {
                actives[a] = b;
                return;
            }
        fprintf(stderr, "Couldn't add an active!\n");
    }
    
    void AddEnemy(const Enemy& b)
    {
        for(unsigned a=0; a<MaxEnemies; ++a)
            if(enemies[a].screen == 255)
            {
                enemies[a] = b;
                return;
            }
        fprintf(stderr, "Couldn't add an enemy!\n");
    }
};

struct LevelPairData
{
    Level& level1;
    Level& level2;
    
    unsigned BeginAddr;
    
    void CombineGFX(bool Combine);
};

#endif
